#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Dec 23 23:26:59 2019

@author: apn
"""
import math


def convert(x, u):
    return x/u


def to(u):
    return 1/u


# factors
T = tera = 1e12
G = giga = 1e9
M = mega = 1e6
K = k = kilo = 1e3
c = centi = 1e-2
m = milli = 1e-3
mu = micro = 1e-6
n = nano = 1e-9
p = pico = 1e-12
f = fempto = 1e-15
a = atto = 1e-18
z = zepto = 1e-21
y = yocto = 1e-24

pi = math.pi

eV = 1
c = 1
kb = 1
hbar = 1

# fundamental
c0 = 299792458  # m/s
hbar0 = 4.135667662e-15/(2*pi)
kb0 = 8.617333262145e-5
# from wikipedia https://de.wikipedia.org/wiki/Nat%C3%BCrliche_Einheiten
joule = 1/(1.60218e-19)
meter = 1/hbar0/c0
s = second = 1/hbar0
gram = 1/kilo*(1/1.78266e-36)  # *c0*c0
kelvin = 1*kb0

# composite
barn = 1e-28*meter**2
u = 1/(6.022141e26)*kilo*gram
Bq = 1/s
Ci = 37*giga * Bq
year = 31556952 * s  # inclusive

# print(convert(1*(y*meter)**2,z*barn))

#print(1e24*u *to(kilo*gram))
#print( (G*eV)**-2 * to(m*barn))



import math
num = dict({'length': -1, 'mass': 1, 'time': -1,
           'temperature': 1, 'momentum': 1, 'energy': 1})
rev_num = dict([reversed(i) for i in num.items()])


class natural_unit:
    def __init__(self, val=1, massdim=1):
        self.value = val
        if(massdim in num):
            massdim = num[massdim]
        self.massdimension = massdim

    def check(self, other):
        assert self.massdimension == other.massdimension

    def __pow__(self, other, modulo=None):
        if isinstance(other, natural_unit):
            assert other.massdimension == 0
            other = other.value
        return natural_unit(self.value**other, self.massdimension*other)

    def __rpow__(self, other, modulo=None):
        assert self.massdimension == 0
        return other**self.value

    # TODO same for subs !!!
    def __radd__(self, other):
        if isinstance(other, natural_unit):
            assert self.massdimension == other.massdimension
            return natural_unit(other.value+self.value, self.massdimension)
        else:
            assert self.massdimension == 0
            return other+self.value, self.massdimension

    def __add__(self, other):
        if isinstance(other, natural_unit):
            assert self.massdimension == other.massdimension
            return natural_unit(self.value + other.value, self.massdimension)
        else:
            assert self.massdimension == 0
            return self.value+other, self.massdimension
    # TODO same for subs !!!

    def __rsub__(self, other):
        assert self.massdimension == other.massdimension
        return natural_unit(other.value - self.value, self.massdimension)

    def __sub__(self, other):
        assert self.massdimension == other.massdimension
        return natural_unit(self.value - other.value, self.massdimension)

    def __rmul__(self, other):
        if isinstance(other, natural_unit):
            ret = natural_unit(other.value * self.value,
                               self.massdimension+other.massdimension)
        else:
            ret = natural_unit(other * self.value, self.massdimension)
        if ret.massdimension == 0:
            return ret.value
        else:
            return ret

    def __mul__(self, other):
        if isinstance(other, natural_unit):
            ret = natural_unit(self.value * other.value,
                               self.massdimension+other.massdimension)
        else:
            ret = natural_unit(self.value * other, self.massdimension)
        if ret.massdimension == 0:
            return ret.value
        else:
            return ret

    def __rtruediv__(self, other):
        if isinstance(other, natural_unit):
            ret = natural_unit(other.value / self.value,
                               other.massdimension-self.massdimension)
        else:
            ret = natural_unit(other / self.value, -self.massdimension)
        if ret.massdimension == 0:
            return ret.value
        else:
            return ret

    def __truediv__(self, other):
        if isinstance(other, natural_unit):
            ret = natural_unit(self.value / other.value,
                               self.massdimension-other.massdimension)
        else:
            ret = natural_unit(self.value / other, self.massdimension)
        if ret.massdimension == 0:
            return ret.value
        else:
            return ret

    def __str__(self):
        if self.massdimension == 0:
            return self.value.__str__()
        return self.value.__str__() + "[" + self.massdimension.__str__() + "]"

    def __repr__(self):
        if self.massdimension == 0:
            return self.value.__repr__()
        return self.value.__repr__() + "[" + self.massdimension.__repr__() + "]"

    def __format__(self, fmt):
        if self.massdimension == 0:
            return self.value.__format__(fmt)
        return self.value.__format__(fmt) + "[" + self.massdimension.__repr__() + "]"


def convert(x, u):
    return x/u


def to(u):
    return 1/u


# factors
T = tera = 1e12
G = giga = 1e9
M = mega = 1e6
K = k = kilo = 1e3
c = centi = 1e-2
m = milli = 1e-3
mu = micro = 1e-6
n = nano = 1e-9
p = pico = 1e-12
f = fempto = 1e-15
a = atto = 1e-18
z = zepto = 1e-21
y = yocto = 1e-24

pi = math.pi

eV = natural_unit(massdim='energy')
c = natural_unit(massdim='length')/natural_unit(massdim='time')
kb = natural_unit(massdim='energy')/natural_unit(massdim='temperature')
hbar = natural_unit(massdim='energy')*natural_unit(massdim='time')


# fundamental
c0 = 299792458 * c  # m/s
hbar0 = 4.135667662e-15/(2*pi)*hbar
kb0 = 8.617333262145e-5*kb

# from wikipedia https://de.wikipedia.org/wiki/Nat%C3%BCrliche_Einheiten
J = joule = 1/(1.60218e-19)*eV

meter = 1/hbar0/c0 * natural_unit(massdim='length')
meter.check(natural_unit(massdim='length'))

s = second = 1/hbar0*natural_unit(massdim='time')
s.check(natural_unit(massdim='time'))

gram = 1/kilo*(1/1.78266e-36) * natural_unit(massdim='mass')
gram.check(natural_unit(massdim='mass'))

# *c0*c0
kelvin = 1*kb0*natural_unit(massdim='temperature')
kelvin.check(natural_unit(massdim='temperature'))

# composite
barn = 1e-28*meter**2
barn.check(natural_unit(massdim='length')**2)

u = 1/(6.022141e26)*kilo*gram
u.check(natural_unit(massdim='mass'))

Bq = 1/s
Ci = 37*giga * Bq
year = 31556952 * s  # inclusive
Hz = hertz = 1/s
W = watt = J/s



from multiprocessing import Process, Queue
from smpl.doc import append_doc


def queued(q, f, *args, **kwargs):
    q.put(f(*args, **kwargs))


def res(a):
    """
        Parallel evaluation of the list generator from :func:`gen`

        Examples
        ========

        >>> def twice(x):
        ...     return x+x
        >>> for r in [calc(twice,i) for i in range(0,5)]:
        ...     print(res(r))
        0
        2
        4
        6
        8
        >>> res([calc(lambda x : x**3, i) for i in range(0,5)])
        [0, 1, 8, 27, 64]

        Return parallel executed values
    """
    if isinstance(a, list):
        return [next(k) for k in a]
    return next(a)


def gen(f, *args, **kwargs):
    """
        Generates parallel execution list generator
    """
    q = Queue()
    p = Process(target=queued, args=(q, f, *args), kwargs=kwargs)
    yield p.start()
    yield [q.get(), p.join()][0]


@append_doc(res)
def calc(f, *args, **kwargs):
    g = gen(f, *args, **kwargs)
    next(g)
    return g


def par(f, *args, **kwargs):
    """
        Parallel execution of f on each element of args and kwargs

        Examples
        ========

        >>> par(lambda x : x**2, range(0,5))
        [0, 1, 4, 9, 16]

    """
    return res([calc(f, *[args[k][i] for k in range(len(args))], **{k: v[i] for k, v in kwargs.items()}) for i in range(len(args[0]) if len(args) > 0 else len(next(iter(kwargs.values()))))])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Simplified python code documentation."""
from .doc import name_to_str, append_doc, append_plot, append_str, insert_doc, insert_str, insert_eq, table, insert_latex, insert_latex_eq
__all__ = ['name_to_str', 'append_doc', 'append_plot', 'append_str', 'insert_doc',
           'insert_str', 'insert_eq', 'table', 'insert_latex', 'insert_latex_eq']



import numpy as np


def si(s, u="", fmt="{}"):
    """
    Get number with uncertainty and unit in ``si`` format for latex.

    Parameters
    ==========
    s : ufloat
        number to be returned in a latex compatible format
    u : str
        unit of that number
    fmt : str
        format string for the numbers

    Returns
    =======
    sistr : str
        latex SI string of the number with it's uncertainty and unit.

    Examples
    ========
    >>> import uncertainties as unc
    >>> from smpl import io
    >>> si(unc.ufloat(2000,0.1))
    '\\\\SI{2000.00+-0.10}{}'
    >>> si(unc.ufloat(2000,0.1),"\\meter")
    '\\\\SI{2000.00+-0.10}{\\\\meter}'
    >>> si(unc.ufloat(2000,0.1),"\\meter", io.gf(2))
    '\\\\SI{2.0+-0.0e+03}{\\\\meter}'

    """
    return "\\SI{%s}{%s}" % ((fmt.format(s)).replace("/", "").replace("(", "").replace(")", ""), u)


def si_line(a, skip=0, fmt="{}"):
    """
    Get array ``a`` in the format of a line of a latex table.

    Examples
    ========
    >>> si_line([1,2,3,])
    '\\\\SI{1}{}&\\\\SI{2}{}&\\\\SI{3}{}\\\\\\\\\\n'

    """
    return si_tab(np.transpose([[t] for t in a]), skip, fmt)


def si_ttab(tab, skip=0, fmt="{}"):
    """
    Transposed :func:`si_tab`.

    Parameters
    ==========
    tab : array_like
        Array containing the values of the table
    skip : number
        Skip this many table lines
    fmt : str
        format string for the numbers

    Returns
    =======
    tabstr : str
        table latex string

    Examples
    ========
    >>> si_ttab([[1,2],[3,4]])
    '\\\\SI{1}{}&\\\\SI{3}{}\\\\\\\\\\n\\\\SI{2}{}&\\\\SI{4}{}\\\\\\\\\\n'

    """
    return si_tab(np.transpose(tab), skip, fmt)


def si_tab(tab, skip=0, fmt="{}"):
    """
    Get arrays of (uncertainty) numbers in  a latex table compatible form.

    Parameters
    ==========
    tab : array_like
        Array containing the values of the table
    skip : number
        Skip this many table lines
    fmt : str
        format string for the numbers

    Returns
    =======
    tabstr : str
        table latex string

    Examples
    ========
    >>> si_tab([[1,2],[3,4]])
    '\\\\SI{1}{}&\\\\SI{2}{}\\\\\\\\\\n\\\\SI{3}{}&\\\\SI{4}{}\\\\\\\\\\n'

    """
    # mkdirs(fn)
    #file = open(fn,"w")
    s = ""
    for _, ti in enumerate(tab):
        for j, tij in enumerate(ti):
            if(j != 0):
                s += "&"
            if(j >= skip):
                s += si(tij, fmt=fmt)
            else:
                s += "%s" % (tij)
        s += "\\\\\n"
    return s



"""A collection of simplified utilities."""

import pkg_resources as pkg  # part of setuptools
#import json
#import requests
#from urllib.request import urlopen

package = "smpl"

try:
    version = pkg.require(package)[0].version
except pkg.DistributionNotFound:
    version = "dirty"

__version__ = version

# def is_internet_available():
#    try:
#        urlopen('http://216.58.192.142', timeout=1)
#        return True
#    except:
#        return False


#
# repository_url='https://pypi.python.org/pypi/%s/json'
#
# url = repository_url % package
#
# if is_internet_available():
#
#     response = requests.get(url).text
#
#     latest_version =  json.loads(response)['info']['version']
#

#
#     parsed_latest = pkg.parse_version(latest_version)
#
#     parsed_version = pkg.parse_version(version)
#

#
#     if parsed_latest > parsed_version:
#
#         print("New version " + str(latest_version) + " > " + str(version) + " available via: \n $ pip install " + package + " --upgrade [--user]'")
#
#     elif parsed_latest < parsed_version:
#
#         print("You are using an unreleased version.")



from smpl import util
from smpl import wrap
params = 0
name = 1


def _append(txt):
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += txt
        return target
    return wrapper


def append(txt):
    """
    TODO split cases doc/txt/plot?
    """
    return None


def append_str(txt):
    return _append(txt)


def append_plot(*args, xmin=-5, xmax=5):
    """
    Append a plot to a function.
    """
    # return _append("\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from " + target.__module__ + " import " +target.__name__ + "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function("+ target.__name__ + "," + ','.join([str(a) for a in args]) + ",xmin="+str(xmin) + ",xmax=" + str(xmax)+")")

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ += "\n\n\t.. plot::\n\t\t:include-source:\n\n\t\t>>> from " + target.__module__ + " import " + target.__name__ + \
            "\n\t\t>>> from smpl import plot\n\t\t>>> plot.function(" + target.__name__ + "," + ','.join(
                [str(a) for a in args]) + ",xmin="+str(xmin) + ",xmax=" + str(xmax)+")"
        # print(target.__doc__)
        return target
    return wrapper


def append_doc(original):
    """
    Append doc string of ``original`` to ``target`` object.

    Parameters
    ==========
    original : ``class`` or ``function``
        ``orignal.__doc__`` is appended to the ``__doc__`` of the ``target``

    Examples
    ========

    >>> def ho():
    ...     '''Ho'''
    ...     print(ho.__doc__)
    >>> @append_doc(ho)
    ... def hi():
    ...     '''Hi'''
    ...     print(hi.__doc__)
    >>> hi()
    HiHo
    """
    return _append(original.__doc__)


def _insert(txt):
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ = txt + target.__doc__
        return target
    return wrapper


def insert_str(txt):
    return _insert(txt)


def insert_doc(original):
    """
    Inserts the docstring from passed function ``original`` in the ``target`` function docstring.
    """
    return _insert(original.__doc__)

    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        target.__doc__ = original.__doc__ + target.__doc__
        return target
    return wrapper


def insert_eq():
    """
    Inserts the function and its parameters and an equal sign
    """
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = ""
        safe = target.__doc__
        target.__doc__ = target.__name__ + "("
        for v in target.__code__.co_varnames:
            target.__doc__ += v + ","
        target.__doc__ = target.__doc__[:-1]
        target.__doc__ += ") = " + safe
        return target
    return wrapper


def insert_latex_eq():
    """
    Inserts latexed code of a oneline function with parameters.
    """
    return lambda f: insert_eq()(insert_latex()(f))


def insert_latex():
    """
    Inserts latexed code of a oneline function.
    """
    def wrapper(target):
        if target.__doc__ is None:
            target.__doc__ = wrap.get_latex(target)
        return target
    return wrapper


tab_len = 20


def table_sep(tabs=1):
    return "="*(tab_len-2) + "  " + "="*(tab_len-2) + "  " + "="*(tab_len-2) + "\n" + "\t".join(["" for i in range(0, tabs+1)])


def table(dic, top=True, bottom=True, init=True, tabs=1):
    """
    Add dict= {'key': [values...]} to a simple reST table.
    """
    t = util.times("\t", tabs)
    rs = ""
    if init:
        rs += t
    if top:
        rs += table_sep(tabs=tabs)
    for k, vs in dic.items():
        rs += str(k)
        p = str(k)
        for v in vs:
            rs += "".join([' ' for i in range(tab_len-len(p))]) + str(v)
            p = str(v)
        rs += "\n" + t
    if bottom:
        rs += table_sep(tabs=tabs)
    else:
        rs += "\n"
    return rs


def name_to_str():
    """
    Replace __str__ with __name__.
    """
    def wrapper(target):
        target.__str__ = lambda s: target.__name__
        return target
    return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod()
